/*******************************************************************************
* Freescale Semiconductor Inc.
* (c) Copyright 2011 Freescale Semiconductor, Inc.
* ALL RIGHTS RESERVED.
********************************************************************************
Services performed by FREESCALE in this matter are performed AS IS and without
any warranty. CUSTOMER retains the final decision relative to the total design
and functionality of the end product. FREESCALE neither guarantees nor will be
held liable by CUSTOMER for the success of this project.
FREESCALE DISCLAIMS ALL WARRANTIES, EXPRESSED, IMPLIED OR STATUTORY INCLUDING,
BUT NOT LIMITED TO, IMPLIED WARRANTY OF MERCHANTABILITY OR FITNESS FOR
A PARTICULAR PURPOSE ON ANY HARDWARE, SOFTWARE ORE ADVISE SUPPLIED 
TO THE PROJECT BY FREESCALE, AND OR NAY PRODUCT RESULTING FROM FREESCALE 
SERVICES. IN NO EVENT SHALL FREESCALE BE LIABLE FOR INCIDENTAL OR CONSEQUENTIAL 
DAMAGES ARISING OUT OF THIS AGREEMENT.
CUSTOMER agrees to hold FREESCALE harmless against any and all claims demands 
or actions by anyone on account of any damage, or injury, whether commercial,
contractual, or tortuous, rising directly or indirectly as a result 
of the advise or assistance supplied CUSTOMER in connection with product, 
services or goods supplied under this Agreement.
********************************************************************************
* File:             main.c
* Owner:            b05111
* Version:          1.9_DPM
* Date:             Jul-28-2015
* Classification:   General Business Information
* Brief:            MPC5643L Pin toggle stationery for DPM mode 
********************************************************************************
* Detailed Description: 
* Application performs basic initialization, setup PLL to maximum allowed freq.,
* initializes interrupts, blinking one LED by interrupt, third LED by software
* loop, initializes and display notice via UART terminal and then terminal ECHO.
* Second core is started and second LED blinking is being performed by it.
* The example configures the device for maximum performance (OPTIMIZATIONS_ON)
* by initialization of instruction cache and enabling of branch prediction.
* Example suppose MCU is configured for DPM (Decoupled-parallel mode). 
* Its intention is to offer advanced startup code additional to CW stationery.
*
* ------------------------------------------------------------------------------
* Test HW:         xPC564xLKIT, PPC5643L Cut3 silicon
* Target :         internal_FLASH, RAM
* Fsys:            120 MHz PLL0
* Debugger:        Lauterbach Trace32
*                  PeMicro USB-ML-PPCNEXUS
* Terminal:        19200-8-no parity-1 stop bit-no flow control via LINFlex0
* EVB connection:  default 
*
********************************************************************************
Revision History:
1.0       Apr-08-2011     b08110  Initial Version
1.1       Apr-12-2011     b08110  Fixed INTC interrupts (IVOR4 init).
                                  Changed header file to Cut2 silicon.
1.2       Jul-08-2011     b08110  Modified Mode Entry sequence in order to enter
                                  a different device mode along with a change
                                  in device clocking.
1.3       Aug-26-2011     b08110  Corrected clk_out settings for output clock
                                  (in the CGM)
1.4       Oct-26-2011     b08110  -Initialized all core registers in __start()
                                  entry point - need for LockStep operation.
                                  -Added Mode transition from SAFE to DRUN
                                  in ClearFails().
                                  -Removed call to ClearCF() after startup.
                                  It is important to review critical faults 
                                  and not to clear them automatically.
                                  -Changed system clock frequency to 120 MHz.
1.5       Jul-29-2014     b05111  Added printf, minor edits
1.6_DPM   Aug-08-2014     b05111  - now based on CW210
                                  - one version for LSM and one for DPM
                                  - added z0 start and blinking LED
                                  - added lauterbach script for switch into DPM
1.9_DPM   Jul-28-2015     b05111  - ModesAndClks_init revised
                                  - added Optimizations
                                  - Lauterbach scripts changed                                                     
*******************************************************************************/

#include <stdio.h>
#include "MPC5643L.h"
#include "uart.h"
#include "IntcInterrupts.h"
#include "Optimizations.h"

/*******************************************************************************
* External objects
*******************************************************************************/

// Prototype for second core startup
extern void __start_p1();

/*******************************************************************************
* Global variables
*******************************************************************************/

/*******************************************************************************
* Constants and macros
*******************************************************************************/

/* Enable caches, branch prediction and optimize flash setting */
#define OPTIMIZATIONS_ON 1

/* FCCU keys */
#define FCCU_NCFK_KEY 0xAB3498FE
#define FCCU_CFK_KEY  0x618B7A50

/* PCR numbers */
#define PD4_pin 52
#define PD6_pin 54
#define PB6_pin 22

#define LED1_pin   PD4_pin
#define LED3_pin   PD6_pin
#define CLKOUT_pin PB6_pin


/*******************************************************************************
* Local types
*******************************************************************************/

/*******************************************************************************
* Local function prototypes
*******************************************************************************/
static void HW_init(void);
static void SysClk_init(void);
static void ModesAndClks_init(void);
static void PeriClkGen_init(void);
static void DisableWatchdog(void);
static void GPIO_init(void);
static void ClearNCF(void);
static void ClearCF(void);
static void ClearFails(void);
static void PIT3_init(void);
static void PIT3_isr(void);
static void Terminal_ECHO(void);

/*******************************************************************************
Function Name : HW_init
Engineer      : b08110
Date          : Jul-28-2015
Parameters    : 
Modifies      : 
Returns       : 
Notes         : initialization of the hw for the purposes of this example
Issues        : 
*******************************************************************************/
static void HW_init(void)
{
    DisableWatchdog();
    ClearFails();

    GPIO_init();    
    SysClk_init();
   
    #if OPTIMIZATIONS_ON
        Global_optimizations();   
        Core_optimizations();
    #endif       
}


/*******************************************************************************
Function Name : SysClk_init
Engineer      : b08110
Date          : Feb-25-2010
Parameters    : 
Modifies      : 
Returns       : 
Notes         : 
Issues        : 
*******************************************************************************/
static void SysClk_init(void)
{
    ModesAndClks_init();
    PeriClkGen_init();
}

/*******************************************************************************
Function Name : ModesAndClks_init
Engineer      : b05111
Date          : Feb-23-2015
Parameters    : 
Modifies      : 
Returns       : 
Notes         : 
Issues        : progressive clock switching can be too fast for EVB regulator
*******************************************************************************/
static void ModesAndClks_init(void) 
{

/*  Used configurations:
            IRCOSC  XOSCON  PLL0    PLL1    sysclk
    DRUN -  ON      OFF     OFF     OFF     IRCOSC
    RUN0 -  ON      ON      OFF     OFF     IRCOSC
    RUN1 -  ON      ON      ON      OFF     IRCOSC
    RUN2 -  ON      ON      ON      OFF     PLL0
    RUN3 -  ON      ON      ON      ON      PLL0
    
    Steps: 
    - DRUN -> RUN0
    - RUN0 -> RUN1
    - wait 200us
    - RUN1 -> RUN2
    - wait 200us
    - RUN2 -> RUN3
    - enable CLKOUT
    - wait 10us
    - decrease ODF for both PLLs is several steps if needed        */

    #define SLOW_MANUAL_PLL_TRANSITION 1 // 0 = progressive clock switching
    
    int32_t cnt = 0;
    
    /* Enable DRUN, RUN0-3, SAFE, RESET modes */
    ME.MER.R = 0x000000FD;
    
    /* Mode Transition to enter RUN0 mode: */
    // RUN0 cfg: 16MHzIRCON,OSC0ON,PLL0OFF,PLL1OFF,syclk=16M IRC
    ME.RUN[0].R = 0x001F0030;
    // Enter RUN0 Mode & Key
    ME.MCTL.R = 0x40005AF0;
    // Enter RUN0 Mode & Inverted Key
    ME.MCTL.R = 0x4000A50F;
    // Wait for mode entry to complete
    while(0 == ME.GS.B.S_XOSC) {};
    // Wait for mode transition to complete
    while(1 == ME.GS.B.S_MTRANS) {}
    // Check RUN0 mode has been entered
    while(4 != ME.GS.B.S_CURRENT_MODE) {};

    /* Initialize PLL before turning it on:
       fsys = fcrystal*ndiv/idf/odf
       fvco must be from 256MHz to 512MHz
       we want fsys = 120MHz. fvco = fsys*odf = 120MHz * 4 = 480MHz
       fsys = 40*72/6/4 = 120MHz */

    CGM.AC3SC.R = 0x01000000; // Select Xosc as PLL0 source clock
    CGM.AC4SC.R = 0x01000000; // Select Xosc as PLL1 source clock

    #if SLOW_MANUAL_PLL_TRANSITION
        /* set ODF=3 at first for smooth transtion */
        /* 120 MHz target - at this moment 30MHz */
        CGM.FMPLL[0].CR.B.IDF = 0x5; // FMPLL0 IDF=5 --> divide by 6
        CGM.FMPLL[0].CR.B.ODF = 0x3; // FMPLL0 ODF=3 --> divide by 16
        CGM.FMPLL[0].CR.B.NDIV = 72; // FMPLL0 NDIV=72 --> divide by 72
        
    #else
        /* 120 MHz */
        CGM.FMPLL[0].CR.B.IDF = 0x5; // FMPLL0 IDF=5 --> divide by 6
        CGM.FMPLL[0].CR.B.ODF = 0x1; // FMPLL0 ODF=1 --> divide by 4
        CGM.FMPLL[0].CR.B.NDIV = 72; // FMPLL0 NDIV=72 --> divide by 72
        CGM.FMPLL[0].CR.B.EN_PLL_SW = 1; // enable progressive clock switching
    #endif
    
    
    ME.RUNPC[0].R = 0x000000FE; // enable peripherals run in all modes
    ME.LPPC[0].R = 0x00000000;  // disable peripherals run in LP modes

    /* Mode Transition to enter RUN1 mode: */
    // RUN1 cfg: 16MHzIRCON,OSC0ON,PLL0ON,PLL1OFF,syclk=16M IRC
    ME.RUN[1].R = 0x001F0070;
    // Enter RUN1 Mode & Key
    ME.MCTL.R = 0x50005AF0;
    // Enter RUN1 Mode & Inverted Key
    ME.MCTL.R = 0x5000A50F;
    // Wait for mode transition to complete
    while (1 == ME.GS.B.S_MTRANS) {}
    // Check RUN1 mode has been entered
    while(5 != ME.GS.B.S_CURRENT_MODE) {};   

    /* wait approx 200us */  
    for(cnt=0;cnt<6000;cnt++) {} 

    /* Mode Transition to enter RUN2 mode: */
    // RUN2 cfg: 16MHzIRCON,OSC0ON,PLL0ON,PLL1OFF,syclk=PLL0
    ME.RUN[2].R = 0x001F0074;
    // Enter RUN2 Mode & Key
    ME.MCTL.R = 0x60005AF0;
    // Enter RUN2 Mode & Inverted Key
    ME.MCTL.R = 0x6000A50F;
    // Wait for mode transition to complete
    while(1 == ME.GS.B.S_MTRANS) {}
    // Check RUN2 mode has been entered
    while(6 != ME.GS.B.S_CURRENT_MODE) {};

    /* wait approx 200us */  
    for(cnt=0;cnt<6000;cnt++) {} 
    
    #if SLOW_MANUAL_PLL_TRANSITION
        /* 80 MHz target - at this moment 20MHz */
        CGM.FMPLL[1].CR.B.IDF = 0x7; // FMPLL0 IDF=5 --> divide by 8
        CGM.FMPLL[1].CR.B.ODF = 0x3; // FMPLL0 ODF=1 --> divide by 16
        CGM.FMPLL[1].CR.B.NDIV = 64; // FMPLL0 NDIV=72 --> divide by 64
    #else
        
        /* 80 MHz */
        CGM.FMPLL[1].CR.B.IDF = 0x7; // FMPLL0 IDF=5 --> divide by 8
        CGM.FMPLL[1].CR.B.ODF = 0x1; // FMPLL0 ODF=1 --> divide by 4
        CGM.FMPLL[1].CR.B.NDIV = 64; // FMPLL0 NDIV=72 --> divide by 64
        CGM.FMPLL[1].CR.B.EN_PLL_SW = 1; // enable progressive clock switching
    #endif
    
    /* Mode Transition to enter RUN3 mode: */
    // RUN3 cfg: 16MHzIRCON,OSC0ON,PLL0ON,PLL1ON,syclk=PLL0
    ME.RUN[3].R = 0x001F00F4;
    // Enter RUN3 Mode & Key
    ME.MCTL.R = 0x70005AF0;
    // Enter RUN3 Mode & Inverted Key
    ME.MCTL.R = 0x7000A50F;
    // Wait for mode transition to complete
    while(1 == ME.GS.B.S_MTRANS) {}
    // Check RUN3 mode has been entered
    while(7 != ME.GS.B.S_CURRENT_MODE) {}; 
    
    /* enable CLKOUT on PB6 */
    SIU.PCR[CLKOUT_pin].R = 0x0600; // ALT1 - PCR[22] - PA = 0b01
    
    /* set CLKOUT divider of 8 */
    CGM.OCDS_SC.R = 0x32000000; // div by 8, system FMPLL, cut2
    CGM.OCEN.B.EN = 1; 			// enable CLKOUT signal
    
    #if SLOW_MANUAL_PLL_TRANSITION
    
        /* wait approx 200us */  
        for(cnt=0;cnt<6000;cnt++) {} 
        CGM.FMPLL[0].CR.B.ODF = 0x2; // FMPLL0 ODF=2 --> divide by 8
        
        /* wait approx 10us */  
        for(cnt=0;cnt<300;cnt++) {} 
        CGM.FMPLL[1].CR.B.ODF = 0x2; // FMPLL0 ODF=2 --> divide by 8
        
        /* wait approx10us */  
        for(cnt=0;cnt<300;cnt++) {}
        // final 120MHz 
        CGM.FMPLL[0].CR.B.ODF = 0x1; // FMPLL0 ODF=1 --> divide by 4
        
        /* wait approx 10us */  
        for(cnt=0;cnt<300;cnt++) {} 
        // final 80MHz
        CGM.FMPLL[1].CR.B.ODF = 0x1; // FMPLL0 ODF=1 --> divide by 4
        
    #endif
}


/*******************************************************************************
Function Name : PeriClkGen_init
Engineer      : b08110
Date          : Apr-14-2010
Parameters    : 
Modifies      : 
Returns       : 
Notes         : - Enable auxiliary clock 0-3. from PLL0
Issues        : 
*******************************************************************************/
static void PeriClkGen_init(void) 
{
    CGM.AC0SC.R = 0x04000000;  /* MPC56xxP: Select PLL0 for aux clk 0  */
    CGM.AC0DC.R = 0x80800000;  /* MPC56xxP: Enable aux clk 0 div by 1 */
    CGM.AC1SC.R = 0x04000000;  /* MPC56xxP: Select PLL0 for aux clk 1  */
    CGM.AC1DC.R = 0x80000000;  /* MPC56xxP: Enable aux clk 1 div by 1 */
    CGM.AC2SC.R = 0x04000000;  /* MPC56xxP: Select PLL0 for aux clk 2  */
    CGM.AC2DC.R = 0x80000000;  /* MPC56xxP: Enable aux clk 2 div by 1 */
}

/*******************************************************************************
Function Name : DisableWatchdog
Engineer      : b08110
Date          : Jun-23-2010
Parameters    : 
Modifies      : 
Returns       : 
Notes         : - disable SWT functionality
Issues        : 
*******************************************************************************/
static void DisableWatchdog(void) 
{
    SWT.SR.R = 0x0000c520;     /* Write keys to clear soft lock bit */
    SWT.SR.R = 0x0000d928; 
    SWT.CR.R = 0x8000010A;     /* Clear watchdog enable (WEN) */
    
    /* e200 Core Watchdog Timer */
    asm
    {
        li  r3, 0
        mtspr   340, r3
    }
}

/*******************************************************************************
Function Name : GPIO_init
Engineer      : b08110
Date          : Feb-08-2010
Parameters    : 
Modifies      : 
Returns       : 
Notes         : 
Issues        : 
*******************************************************************************/
static void GPIO_init(void)
{    	 	
 	SIU.GPDO[LED1_pin].R = 1;
    SIU.GPDO[LED3_pin].R = 1;
    
    SIU.PCR[LED1_pin].B.PA = 0x00; 
    SIU.PCR[LED3_pin].B.PA = 0x00;
    
    SIU.PCR[LED1_pin].B.OBE = 0x01;
    SIU.PCR[LED3_pin].B.OBE = 0x01;    
    
}

/*******************************************************************************
Function Name : ClearNCF
Engineer      : b08110
Date          : Apr-08-2011
Parameters    : 
Modifies      : 
Returns       : 
Notes         : 
Issues        : 
*******************************************************************************/
static void ClearNCF(void)
{
    uint32_t i,b[4];
    for(i=0;i<4;i++)
    {
        FCCU.NCFK.R = FCCU_NCFK_KEY;
        FCCU.NCF_S[i].R = 0xFFFFFFFF;
        while(FCCU.CTRL.B.OPS != 0x3)
        {
        
        };              /* wait for the completion of the operation */
        b[i]=FCCU.NCF_S[i].R;
    }
}

/*******************************************************************************
Function Name : ClearCF
Engineer      : b08110
Date          : Apr-08-2011
Parameters    : 
Modifies      : 
Returns       : 
Notes         : 
Issues        : 
*******************************************************************************/
static void ClearCF(void)
{
    uint32_t i,a[4];
    for(i=0;i<4;i++)
    {
        FCCU.CFK.R = FCCU_CFK_KEY;
   
        FCCU.CF_S[i].R = 0xFFFFFFFF;
  
        while(FCCU.CTRL.B.OPS != 0x3)
        {
            
        };      /* wait for the completion of the operation */

        a[i]=FCCU.CF_S[i].R;
    }
}

/*******************************************************************************
Function Name : ClearFails
Engineer      : b08110
Date          : Oct-26-2011
Parameters    : 
Modifies      : 
Returns       : 
Notes         : 
Issues        : 
*******************************************************************************/
static void ClearFails(void)
{
    uint16_t reset_register;
    
    if(RGM.FES.B.F_FCCU_SAFE /*|| RGM.FES.B.F_FCCU_HARD*/)
    {
        reset_register = RGM.FES.R;
        ME.IMTS.R = 0x00000001;
        /* ClearCF(); */
        ClearNCF();
        
        RGM.FES.R = 0xFFFF;
        RGM.DES.R = 0xFFFF;
        
        /* re-enter DRUN */
        ME.MCTL.R = 0x30005AF0; /* Enter DRUN Mode & Key */        
        ME.MCTL.R = 0x3000A50F; /* Enter DRUN Mode & Inverted Key */
    }
}


/*******************************************************************************
Function Name : PIT3_init
Engineer      : b08110
Date          : Mar-12-2011
Parameters    : 
Modifies      : 
Returns       : 
Notes         : - 1 second periodic interrup using the PIT3 module
Issues        : - expecting 120 MHz fsys for the PIT timer
*******************************************************************************/
static void PIT3_init(void)
{
	/* 30: MDIS = 0 to enable clock for PITs. */
    /* 31: FRZ = 1 for Timers stopped in debug mode */
    PIT.PITMCR.R = 0x00000001;    
    PIT.LDVAL3.R = 120000000 - 1;
    	
    /* clear the TIF flag */
    PIT.TFLG3.R = 0x00000001;
    	
    /* 30: TIE = 1 for interrupt request enabled */
    /* 31: TEN = 1 for timer active */
    PIT.TCTRL3.R = 0x00000003;
}

/*******************************************************************************
Function Name : PIT3_isr
Engineer      : b08110
Date          : Mar-12-2011
Parameters    : 
Modifies      : 
Returns       : 
Notes         : - 1 second periodic interrupt using the PIT3 module
Issues        : 
*******************************************************************************/
static void PIT3_isr(void)
{
    /* toggle LED */
    SIU.GPDO[LED3_pin].R ^= 1;

    PIT.TFLG3.R = 0x00000001;    
}


/*******************************************************************************
Function Name : Terminal_ECHO
Engineer      : b05111
Date          : Apr-19-2012
Parameters    : 
Modifies      : 
Returns       : 
Notes         :                          
Issues        : ReadUARTPoll must be defined in uart.c
*******************************************************************************/

static void Terminal_ECHO(void)
{
    vuint32_t result;
    char ch; 
    
    /* if character is present, read it and re-sent it to terminal */
    result = ReadUARTPoll(&ch);
    if (result==kUARTNoError) 
    {
        switch(ch)
        {
            // enter
            case '\n':
            case '\r':
                printf("\r\n");
                break;
                    
            // other symbols
            default:
                printf("%c",ch); 
                fflush(stdout);
        }           
    } 

}


/*******************************************************************************
* Global functions
*******************************************************************************/

int32_t main(void) 
{
    vint32_t counter = 0;

    HW_init(); 
    
    /* Install PIT3_isr into the interrupt vector table */
    INTC_InstallINTCInterruptHandler(PIT3_isr, 127, 1);
    
    /* lower current INTC priority to start INTC interrupts */
    INTC.CPR_PRC0.R = 0;
    
    /* 1 sec periodical interrupt */
    PIT3_init();
    
    /* Start the second core, VLE mode*/
    SSCM.DPMBOOT.R = (unsigned long)__start_p1 + 0x00000002; 
    SSCM.DPMKEY.R = 0x00005AF0;
    SSCM.DPMKEY.R = 0x0000A50F;  

    printf("Hello World\r\n");   
    printf("Best Regards, Technical Information Center\r\n");
    
    SIU.GPDO[LED1_pin].R = 0;
    
    /* Loop forever */
    while (1) 
    {
        /* delay */
        while(counter < 500000)
        {
            counter++;
            Terminal_ECHO();

        }  
        counter = 0;
        
        /* toggle LED */
        SIU.GPDO[LED1_pin].R ^= 1;
           
    }
    
}

